import json
from pathlib import Path

from typing import Callable, Union

import pytz
from bs4 import BeautifulSoup
from bs4.element import Tag, NavigableString
from datetime import datetime
from typing import Protocol, TypeAlias
from collections import defaultdict

from jinja2 import Environment

ResourceExtractor: TypeAlias = Callable[[str], tuple[str, list]]


def get_corrects_and_incorrects(question_tag):
    corrects = question_tag.css.filter('correct')
    incorrects = question_tag.css.filter('incorrect')
    return corrects, incorrects


def get_correct_comments(question_tag):
    feedback = question_tag.css.filter('correct-comments')
    return get_text_contents(feedback[0]) if feedback else None


def get_incorrect_comments(question_tag):
    feedback = question_tag.css.filter('incorrect-comments')
    return get_text_contents(feedback[0]) if feedback else None


def get_points(question_tag, default=1):
    points = question_tag.get("points", default)
    try:
        return int(points)
    except ValueError:
        print("Invalid points value: " + points)
        return default


def get_answers(question_tag):
    corrects, incorrects = get_corrects_and_incorrects(question_tag)
    return corrects + incorrects


def string_is_date(date: str):
    # For templating. The string might not be a date yet.
    # Once the template arguments are filled in, we will apply make_iso.
    if date.startswith("{") or "due" in date.lower() or "lock" in date.lower():
        return False
    has_digit = False
    for d in range(10):
        if f"{d}" in date:
            has_digit = True
    return has_digit


def make_iso(date: datetime | str | None, time_zone: str) -> str:
    if isinstance(date, datetime):
        return datetime.isoformat(date)
    elif isinstance(date, str):
        # Check if the string is already in ISO format
        try:
            return datetime.isoformat(datetime.fromisoformat(date))
        except ValueError:
            pass

        try_formats = [
            "%b %d, %Y, %I:%M %p",
            "%b %d %Y %I:%M %p",
            "%Y-%m-%dT%H:%M:%S%z"
        ]
        for format_str in try_formats:
            try:
                parsed_date = datetime.strptime(date, format_str)
                break
            except ValueError:
                pass
        else:
            raise ValueError(f"Invalid date format: {date}")

        # Convert the parsed datetime object to the desired timezone
        to_zone = pytz.timezone(time_zone)
        parsed_date = parsed_date.replace(tzinfo=None)  # Remove existing timezone info
        parsed_date = parsed_date.astimezone(to_zone)
        return datetime.isoformat(parsed_date)
    else:
        raise TypeError("Date must be a datetime object or a string")


def get_text_contents(tag, children_tag_names: list[str] = ()):
    """
    Typically, the body of a tag is found at contents[0], and sub-tags (like answers) are found later.
    However, images sometimes separate the text into multiple parts.
    This function joins the text and images together.
    """
    return "".join(
        [str(c) for c in tag.contents if isinstance(c, NavigableString) or (isinstance(c, Tag) and c.name not in children_tag_names)])


question_types = [
    'calculated_question',
    'essay_question',
    'file_upload_question',
    'fill_in_multiple_blanks_question',
    'matching_question',
    'multiple_answers_question',
    'multiple_choice_question',
    'multiple_dropdowns_question',
    'numerical_question',
    'short_answer_question',
    'text_only_question',
    'true_false_question'
]


class TFConverter:
    @staticmethod
    def process(correct_incorrect_tag, markdown_processor: ResourceExtractor):
        is_true = correct_incorrect_tag.name == "correct"
        question_text, resources = markdown_processor(get_text_contents(correct_incorrect_tag, ["correct-comments", "incorrect-comments"]))
        question = {
            "question_text": question_text,
            "question_type": 'true_false_question',
            "points_possible": get_points(correct_incorrect_tag),
            "correct_comments": get_correct_comments(correct_incorrect_tag),
            "incorrect_comments": get_incorrect_comments(correct_incorrect_tag),
            "answers": [
                {
                    "answer_text": "True",
                    "answer_weight": 100 if is_true else 0
                },
                {
                    "answer_text": "False",
                    "answer_weight": 0 if is_true else 100
                }
            ]
        }
        return question, resources


class Processor(Protocol):
    """
    A processor takes a question tag and a Markdown processor
      returning question(s) and a list of resources
    """

    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor) -> Union[
        tuple[list[dict], list], tuple[dict, list]]: ...


class AttributeAdder:
    def __init__(self, settings: dict, settings_tag: Tag):
        self.settings = settings
        self.settings_tag = settings_tag

    def __call__(self, attribute, default=None, new_name=None, formatter=None):
        if attribute in self.settings_tag.attrs or default is not None:
            value = self.settings_tag.get(attribute, default)
            if formatter:
                value = formatter(value)
            self.settings[new_name if new_name else attribute] = value


def process(processor: Processor, question_tag, markdown_processor: ResourceExtractor):
    question, resources = processor.process(question_tag, markdown_processor)
    return question, resources


class TrueFalseProcessor:
    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor):
        answers = get_answers(question_tag)

        check_correct_size(answers, 1, "True/False")

        question, resources = process(TFConverter(), answers[0], markdown_processor)
        if not get_points(answers[0], 0):
            points = get_points(question_tag)
            question["points_possible"] = points
        if not question["correct_comments"]:
            question["correct_comments"] = get_correct_comments(question_tag)
        if not question["incorrect_comments"]:
            question["incorrect_comments"] = get_incorrect_comments(question_tag)
        return question, resources


class MultipleTrueFalseProcessor:
    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor):
        heading_question, resources = process(TextQuestionProcessor(), question_tag, markdown_processor)
        questions = [heading_question]
        for answer in get_answers(question_tag):
            tf_question, res = process(TFConverter(), answer, markdown_processor)
            questions.append(tf_question)
            resources.extend(res)
        return questions, resources


def check_correct_size(corrects: list, num, question_type):
    if num is not None and len(corrects) != num:
        raise Exception(f"{question_type} must have exactly {num} correct answer(s)\n"
                        "Answers: " + str(corrects))


class MultipleCommonProcessor:
    question_type: str
    num_correct: Union[int, None]

    def process(self, question_tag, markdown_processor: ResourceExtractor):
        corrects, incorrects = get_corrects_and_incorrects(question_tag)
        check_correct_size(corrects, self.num_correct, self.question_type)

        question_text, resources = markdown_processor(get_text_contents(question_tag, ["correct", "incorrect", "correct-comments", "incorrect-comments"]))
        answers = []
        for answer in corrects + incorrects:
            answer_html, res = markdown_processor(get_text_contents(answer))
            answers.append((True if answer in corrects else False, answer_html))
            resources.extend(res)

        question = {
            "question_text": question_text,
            "question_type": self.question_type,
            "points_possible": get_points(question_tag),
            "correct_comments": get_correct_comments(question_tag),
            "incorrect_comments": get_incorrect_comments(question_tag),
            "answers": [
                {
                    "answer_html": answer_html,
                    "answer_weight": 100 if correct else 0
                } for correct, answer_html in answers
            ]
        }
        return question, resources


class MultipleChoiceProcessor(MultipleCommonProcessor):
    question_type = 'multiple_choice_question'
    num_correct = 1

    def process(self, question_tag, markdown_processor: ResourceExtractor):
        return super().process(question_tag, markdown_processor)


class MultipleAnswersProcessor(MultipleCommonProcessor):
    question_type = 'multiple_answers_question'
    num_correct = None

    def process(self, question_tag, markdown_processor: ResourceExtractor):
        return super().process(question_tag, markdown_processor)


class MatchingProcessor:
    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor):
        pairs = question_tag.css.filter('pair')
        matches = []
        for pair in pairs:
            answer_left, answer_right = pair.css.filter('left')[0], pair.css.filter('right')[0]
            matches.append((answer_left.string.strip(), answer_right.string.strip()))

        distractors = question_tag.css.filter('distractors')
        distractor_text = get_text_contents(distractors[0]).strip() if len(distractors) > 0 else None
        question_text, resources = markdown_processor(get_text_contents(question_tag, ["pair", "correct-comments", "incorrect-comments"]))
        question = {
            "question_text": question_text,
            "question_type": 'matching_question',
            "points_possible": get_points(question_tag),
            "correct_comments": get_correct_comments(question_tag),
            "incorrect_comments": get_incorrect_comments(question_tag),
            "answers": [
                {
                    "answer_match_left": answer_left,
                    "answer_match_right": answer_right,
                    "answer_weight": 100
                } for answer_left, answer_right in matches
            ],
            "matching_answer_incorrect_matches": distractor_text
        }
        return question, resources


class TextQuestionProcessor:
    @staticmethod
    def process(question_tag, markdown_processor: ResourceExtractor):
        question_text, resources = markdown_processor(get_text_contents(question_tag))
        question = {
            "question_text": question_text,
            "question_type": 'text_only_question',
        }
        return question, resources


class Parser:
    @staticmethod
    def get_list(string):
        items = string.strip().split(',')
        return [cell.strip() for cell in items if cell.strip()]

    @staticmethod
    def get_bool(string):
        # Forgiving boolean parser
        if isinstance(string, bool):
            return string

        if string.lower() == "true":
            return True
        elif string.lower() == "false":
            return False
        else:
            raise ValueError(f"Invalid boolean value: {string}")

    @staticmethod
    def get_dict(string):
        # Assumes the string is a comma-separated list of key-value pairs
        # Example: "key1=value1, key2=value2 "
        return dict(cell.strip().split('=') for cell in string.split(',') if cell.strip())


class OverrideParser:
    def __init__(self, date_formatter, templater, parser):
        self.date_formatter = date_formatter
        self.template = templater
        self.parser = parser

    def parse(self, override_tag: Tag):
        override = {
            "type": "override",
            "settings": {},
            "sections": [],
            "students": [],
            "assignments": []
        }
        for tag in override_tag.find_all():
            if tag.name == "section":
                override["sections"].append(get_text_contents(tag))
            elif tag.name == "student":
                override["students"].append(get_text_contents(tag))
            elif tag.name == "assignment":
                override["assignments"].append(self.parse_assignment_tag(tag))
            elif tag.name == "template-arguments":
                override["replacements"] = self.template(tag)
        return override

    def parse_assignment_tag(self, tag):
        settings = {
            "title": tag["title"],
        }
        adder = AttributeAdder(settings, tag)
        adder("available_from", new_name="unlock_at", formatter=self.date_formatter)
        adder("due_at", formatter=self.date_formatter)
        adder("available_to", new_name="lock_at", formatter=self.date_formatter)
        return settings


class ModuleParser:
    def __init__(self, parser):
        self.parser = parser

    def parse_module_settings(self, module_tag):
        settings = {
            "name": module_tag["title"],
            "position": module_tag["position"],
        }
        AttributeAdder(settings, module_tag)("published", False, formatter=self.parser.get_bool)
        return settings

    def parse(self, module_tag: Tag):
        module = {
            "type": "module",
            "name": module_tag["title"],
            "settings": self.parse_module_settings(module_tag),
            "items": []
        }
        for item_tag in module_tag.find_all():
            module["items"].append(self.parse_module_item(item_tag))
        return module

    casing = {
        "file": "File",
        "page": "Page",
        "discussion": "Discussion",
        "assignment": "Assignment",
        "quiz": "Quiz",
        "subheader": "SubHeader",
        "externalurl": "ExternalUrl",
        "externaltool": "ExternalTool"
    }

    def parse_module_item(self, tag: Tag):
        item = {
            "title": tag["title"],
            "type": self.casing[tag.name],
        }

        adder = AttributeAdder(item, tag)
        adder("position")
        adder("indent")
        adder("page_url")
        adder("external_url")
        adder("new_tab", True, formatter=self.parser.get_bool)
        adder("completion_requirement")
        adder("iframe")
        adder("published", False, formatter=self.parser.get_bool)
        return item


class QuizParser:
    question_processors = {
        "multiple-choice": MultipleChoiceProcessor(),
        "multiple-answers": MultipleAnswersProcessor(),
        "true-false": TrueFalseProcessor(),
        "multiple-tf": MultipleTrueFalseProcessor(),
        "matching": MatchingProcessor(),
        "text": TextQuestionProcessor()
    }

    def __init__(self, markdown_processor: ResourceExtractor, group_indexer, date_formatter, templater, parser):
        self.markdown_processor = markdown_processor
        self.group_indexer = group_indexer
        self.date_formatter = date_formatter
        self.template = templater
        self.parser = parser

    def parse(self, quiz_tag: Tag):
        quiz = {
            "type": "quiz",
            "questions": [],
            "resources": [],
            "replacements": [],
            "settings": {}
        }
        for tag in quiz_tag.find_all():
            if tag.name == "settings":
                quiz["settings"] = self.parse_quiz_settings(tag)
                quiz["name"] = quiz["settings"]["title"]
            elif tag.name == "template-arguments":
                quiz["replacements"] = self.template(tag)
            elif tag.name == "question":
                question, res = self.parse_question(tag)
                quiz["resources"].extend(res)
                # if question is a  list of questions, add them all
                if isinstance(question, list):
                    quiz["questions"].extend(question)
                else:
                    quiz["questions"].append(question)
            elif tag.name == "description":
                description, res = self.markdown_processor(get_text_contents(tag))
                quiz["resources"].extend(res)
                quiz["settings"]["description"] = description
        return quiz

    def parse_quiz_settings(self, settings_tag):
        settings = {"title": settings_tag["title"]}

        adder = AttributeAdder(settings, settings_tag)

        adder("quiz_type", "assignment")
        adder("assignment_group", None, "assignment_group_id", formatter=self.group_indexer)
        adder("time_limit")
        adder("points_possible")
        adder("shuffle_answers", False, formatter=self.parser.get_bool)
        adder("hide_results", formatter=self.parser.get_bool)
        adder("show_correct_answers", True, formatter=self.parser.get_bool)
        adder("show_correct_answers_last_attempt", False, formatter=self.parser.get_bool)
        adder("show_correct_answers_at", None, formatter=self.date_formatter)
        adder("hide_correct_answers_at", None, formatter=self.date_formatter)
        adder("allowed_attempts")
        adder("scoring_policy", "keep_highest")
        adder("one_question_at_a_time", False, formatter=self.parser.get_bool)
        adder("cant_go_back", False, formatter=self.parser.get_bool)
        adder("available_from", None, "unlock_at", formatter=self.date_formatter)
        adder("due_at", None, formatter=self.date_formatter)
        adder("available_to", None, "lock_at", formatter=self.date_formatter)
        adder("access_code")
        adder("published", False, formatter=self.parser.get_bool)
        adder("one_time_results", False, formatter=self.parser.get_bool)

        return settings

    def parse_question(self, question_tag: Tag):
        processor = self.question_processors[question_tag["type"]]
        return processor.process(question_tag, self.markdown_processor)


class AssignmentParser:
    def __init__(self, markdown_processor: ResourceExtractor, group_indexer, date_formatter, templater, parser):
        self.markdown_processor = markdown_processor
        self.group_indexer = group_indexer
        self.date_formatter = date_formatter
        self.template = templater
        self.parser = parser

    def parse(self, assignment_tag):
        assignment = {
            "name": "",
            "type": "assignment",
            "resources": [],
            "replacements": [],
            "settings": {}
        }
        for tag in assignment_tag.find_all():
            if tag.name == "settings":
                settings = self.parse_assignment_settings(tag)
                assignment["settings"].update(settings)
            elif tag.name == "template-arguments":
                assignment["replacements"] = self.template(tag)
            elif tag.name == "description":
                contents = get_text_contents(tag)
                description, res = self.markdown_processor(contents)
                assignment["settings"]["description"] = description
                assignment["resources"].extend(res)

        assignment["name"] = assignment["settings"]["name"]
        return assignment

    def parse_assignment_settings(self, settings_tag):
        settings = {"name": settings_tag["title"]}

        adder = AttributeAdder(settings, settings_tag)
        adder("position")
        adder("submission_types", "none", formatter=self.parser.get_list)
        adder("allowed_extensions", "", formatter=self.parser.get_list)
        adder("turnitin_enabled", False, formatter=self.parser.get_bool)
        adder("vericite_enabled", False, formatter=self.parser.get_bool)
        adder("turnitin_settings")
        adder("integration_data")
        adder("peer_reviews", False, formatter=self.parser.get_bool)
        adder("automatic_peer_reviews", False, formatter=self.parser.get_bool)
        adder("notify_of_update", False, formatter=self.parser.get_bool)
        adder("group_category", new_name="group_category_id")
        adder("grade_group_students_individually", False, formatter=self.parser.get_bool)
        adder("external_tool_tag_attributes", "", formatter=self.parser.get_dict)
        adder("points_possible")
        adder("grading_type", "points")
        adder("available_from", new_name="unlock_at", formatter=self.date_formatter)
        adder("due_at", formatter=self.date_formatter)
        adder("available_to", new_name="lock_at", formatter=self.date_formatter)
        adder("assignment_group", new_name="assignment_group_id", formatter=self.group_indexer)
        adder("assignment_overrides")
        adder("only_visible_to_overrides", False, formatter=self.parser.get_bool)
        adder("published", False, formatter=self.parser.get_bool)
        adder("grading_standard_id")
        adder("omit_from_final_grade", False, formatter=self.parser.get_bool)
        adder("hide_in_gradebook", False, formatter=self.parser.get_bool)
        adder("quiz_lti")
        adder("moderated_grading", False, formatter=self.parser.get_bool)
        adder("grader_count")
        adder("final_grader_id")
        adder("grader_comments_visible_to_graders", False, formatter=self.parser.get_bool)
        adder("graders_anonymous_to_graders", False, formatter=self.parser.get_bool)
        adder("grader_names_visible_to_final_grader", False, formatter=self.parser.get_bool)
        adder("anonymous_grading", False, formatter=self.parser.get_bool)
        adder("allowed_attempts", formatter=lambda x: -1 if x == "not_graded" else x)
        adder("annotatable_attachment_id")

        return settings


class PageParser:
    def __init__(self, markdown_processor: ResourceExtractor, date_formatter, parser):
        self.markdown_processor = markdown_processor
        self.date_formatter = date_formatter
        self.parser = parser

    def parse_page_settings(self, page_tag):
        settings = {
            "type": "page",
            "name": page_tag["title"],
            "body": "",
        }
        adder = AttributeAdder(settings, page_tag)
        adder("editing_roles", "teachers")
        adder("notify_of_update", False, formatter=self.parser.get_bool)
        adder("published", False, formatter=self.parser.get_bool)
        adder("front_page", False, formatter=self.parser.get_bool)
        adder("publish_at", formatter=self.date_formatter)
        return settings

    def parse(self, page_tag):
        page = {
            "type": "page",
            "name": page_tag["title"],
            "settings": self.parse_page_settings(page_tag),
            "resources": []
        }
        contents = get_text_contents(page_tag)
        body, res = self.markdown_processor(contents)
        page["settings"]["body"] = body
        page["resources"].extend(res)
        return page


class DocumentParser:
    def __init__(self, path_to_resources: Path, path_to_canvas_files: Path, markdown_processor: ResourceExtractor,
                 time_zone: str,
                 group_identifier=lambda x: 0):
        self.path_to_resources = path_to_resources
        self.path_to_files = path_to_canvas_files
        self.markdown_processor = markdown_processor
        self.date_formatter = lambda x: make_iso(x, time_zone)

        self.jinja_env = Environment()
        # This enables us to use the zip function in template documents

        self.jinja_env.globals.update(zip=zip, split_list=lambda sl: [s.strip() for s in sl.split(';')])

        parser = Parser()

        self.element_processors = {
            "quiz": QuizParser(self.markdown_processor, group_identifier, self.date_formatter,
                               self.parse_template_data, parser),
            "assignment": AssignmentParser(self.markdown_processor, group_identifier, self.date_formatter,
                                           self.parse_template_data, parser),
            "page": PageParser(self.markdown_processor, self.date_formatter, parser),
            "module": ModuleParser(parser),
            "override": OverrideParser(self.date_formatter, self.parse_template_data, parser)
        }

    def parse(self, text):
        soup = BeautifulSoup(text, "html.parser")
        document = []
        tag: Tag
        for tag in soup.children:
            parser = self.element_processors.get(tag.name, None)
            if parser:
                elements = parser.parse(tag)
                if not isinstance(elements, list):
                    elements = [elements]
                for element in elements:
                    new_elements = self.create_elements_from_template(element)
                    document.extend(new_elements)
        return document

    def create_elements_from_template(self, element_template):
        if not (all_replacements := element_template.get("replacements", None)):
            return [element_template]

        # Element template is an object, turn it into text
        template_text = json.dumps(element_template, indent=4)

        # Use the text to create a jinja template
        template = self.jinja_env.from_string(template_text)

        elements = []
        for context in all_replacements:
            for key, value in context.items():
                context[key] = value.replace('"', '\\"')
            # For each replacement, create an object from the template
            rendered = template.render(context)
            element = json.loads(rendered)
            elements.append(element)

        # Replacements become unnecessary after creating the elements
        for element in elements:
            del element["replacements"]
        return elements

    def parse_template_data(self, template_tag):
        """
        Parses a template tag into a list of dictionaries
        Each dictionary will become a canvas object
        Converts the following:
        | header1 | header2    |
        |---------|------------|
        | first   | quiz       |
        | second  | assignment |
        into
        [
            {
                "header1": "first",
                "header2": "quiz"
            },
            {
                "header1": "second",
                "header2": "assignment"
            }
        ]
        """
        if template_tag.get("filename"):
            csv = (self.path_to_files / template_tag.get("filename")).read_text()
            headers, *lines = csv.split('\n')
        else:
            headers, separator, *lines = get_text_contents(template_tag).strip().split('\n')
            # Remove whitespace and empty headers
            headers = [h.strip() for h in headers.split('|') if h.strip()]
            lines = [line for left_bar, *line, right_bar in [line.split('|') for line in lines]]

        data = []
        for line in lines:
            line = [phrase.strip() for phrase in line]

            replacements = defaultdict(dict)
            for header, value in zip(headers, line):
                replacements[header] = value

            data.append(replacements)
        return data



import argparse

from pathlib import Path
from mdxcanvas import post_document, get_course
import json
import os


def main(api_url, api_token, course_id, time_zone: str, file_path: Path):
    print("-" * 50 + "\nCanvas MDX\n" + "-" * 50)

    post_document(get_course(api_url, api_token, course_id), time_zone, file_path)


def entry():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", type=Path)
    parser.add_argument("--course-info", type=Path, default="canvas_course_info.json")
    args = parser.parse_args()

    with open(args.course_info) as f:
        course_settings = json.load(f)

    main(api_url=course_settings["CANVAS_API_URL"],
         api_token=os.getenv("CANVAS_API_TOKEN"),
         course_id=course_settings["CANVAS_COURSE_ID"],
         time_zone=course_settings["LOCAL_TIME_ZONE"],
         file_path=args.filename)


if __name__ == '__main__':
    entry()



import re
import xml.etree.ElementTree as etree

from markdown.inlinepatterns import BacktickInlineProcessor, BACKTICK_RE
from markdown.extensions import Extension


class BlackInlineCodeProcessor(BacktickInlineProcessor):
    def handleMatch(self, m: re.Match[str], data: str) -> tuple[etree.Element | str, int, int]:
        el, start, end = super().handleMatch(m, data)
        el.attrib['style'] = 'color: #000000'
        return el, start, end


class BlackInlineCodeExtension(Extension):
    def extendMarkdown(self, md):
        # We use 'backtick' and 190 which are the same values
        # used in markdown.inlinepatterns.py to register the original
        # BacktickInlineCodeProcessor.
        # By reusing the same name, it overrides the original processor with ours
        md.inlinePatterns.register(BlackInlineCodeProcessor(BACKTICK_RE), 'backtick', 190)




import re
from datetime import datetime
import json
import os
import textwrap

import uuid
import argparse
from xml.etree import ElementTree as etree

import pytz
from canvasapi import Canvas
from canvasapi.assignment import Assignment

from canvasapi.quiz import Quiz
from canvasapi.course import Course
from canvasapi.module import Module

import markdown as md
from markdown.extensions.codehilite import makeExtension as makeCodehiliteExtension

from pathlib import Path
from bs4 import BeautifulSoup

from .extensions import BlackInlineCodeExtension
from .parser import DocumentParser, make_iso


def readfile(filepath: Path):
    with open(filepath) as file:
        return file.read()


def get_fancy_html(markdown_or_file: str, files_folder=None):
    """
    Converts markdown to html, and adds syntax highlighting to code blocks.
    """
    if markdown_or_file.endswith('.md'):
        markdown_or_file = readfile(files_folder / markdown_or_file)

    dedented = textwrap.dedent(markdown_or_file)

    fenced = md.markdown(dedented, extensions=[
        'fenced_code',

        # This embeds the highlight style directly into the HTML
        # instead of using CSS classes
        makeCodehiliteExtension(noclasses=True),

        # This forces the color of inline code to be black
        # as a workaround for Canvas's super-ugly default red :P
        BlackInlineCodeExtension()
    ])
    return fenced


def print_red(string):
    print(f"\033[91m{string}\033[00m")


def get_canvas_folder(course: Course, folder_name: str, parent_folder_path=""):
    """
    Retrieves an object representing a digital folder in Canvas. If the folder does not exist, it is created.
    """
    folders = list(course.get_folders())
    if not any(fl.name == folder_name for fl in folders):
        print(f"Created {folder_name} folder")
        course.create_folder(name=folder_name, parent_folder_path=parent_folder_path, hidden=True)
    return [fl for fl in folders if fl.name == folder_name][0]


def create_resource_folder(course, quiz_title: str, course_folders):
    """
    Creates a folder in Canvas to store images and other resources.
    """
    generated_folder_name = "Generated-Content"
    if not any(fl.name == generated_folder_name for fl in course_folders):
        print("Created Content Folder")
        course.create_folder(name=generated_folder_name, parent_folder_path="", hidden=True)

    if not any(fl.name == quiz_title for fl in course_folders):
        print(f"Created {quiz_title} folder")
        course.create_folder(name=quiz_title, parent_folder_path=generated_folder_name,
                             hidden=True)


def get_img_html(image_name, alt_text, style, course, image_folder: Path):
    """
    Returns the html for an image, and the path to the resource, so it can later be uploaded to Canvas.
    After uploading, the correct resource id must be substituted for the fake id using a text replace.
    """
    fake_object_id = str(uuid.uuid4())
    style_text = f'style="{style}"' if style else ""
    html_text = f'<p><img id="{image_name}" src="/courses/{course.id}/files/{fake_object_id}/preview" alt="{alt_text}" {style_text}/></p>'
    resource = (fake_object_id, str(image_folder / image_name))
    return html_text, resource


def process_images(html, course: Course, image_folder):
    """
    Finds all the images in the html, and replaces them with html that links to the image in Canvas.
    Returns the new html, and a list of resources that need to be uploaded.
    After uploading, the correct object id must be substituted for the fake ids, which are returned in the resources.
    """
    soup = BeautifulSoup(html, "html.parser")
    matches = soup.find_all('img')
    resources = []
    for img in matches:
        basic_image_html, resource = get_img_html(img["src"], img["alt"], img.get("style"), course, image_folder)
        img.replace_with(BeautifulSoup(basic_image_html, "html.parser"))
        resources.append(resource)
    return str(soup), resources


def process_markdown(markdown_or_file: str, course: Course, image_folder, files_folder=None):
    """
    Converts markdown to html, and adds syntax highlighting to code blocks.
    Then, finds all the images in the html, and replaces them with html that links to the image in Canvas.
    """
    html = get_fancy_html(markdown_or_file, files_folder)
    return process_images(html, course, image_folder)


def get_group_id(course: Course, group_name: str, names_to_ids: dict[str, int]):
    """
    Group ids are numbers that stand for groups like Labs, Projects, etc.
    Since users will provide names for groups, this method is necessary to find ids.
    """
    if not group_name:
        return None

    if group_name not in names_to_ids:
        print("Created Assignment Group: " + group_name)
        course.create_assignment_group(name=group_name)
        for g in course.get_assignment_groups():
            if g.name not in names_to_ids:
                names_to_ids[g.name] = g.id

    return names_to_ids[group_name]


def replace_questions(quiz: Quiz, questions: list[dict]):
    """
    Deletes all questions in a quiz, and replaces them with new questions.
    """
    print(f"Replacing questions ... ", end="")
    for quiz_question in quiz.get_questions():
        quiz_question.delete()
    for question in questions:
        quiz.create_question(question=question)


def get_section_id(course: Course, section_name: str):
    sections = course.get_sections()
    for section in sections:
        if section.name == section_name:
            return section.id
    print(f"Valid section names: {[s.name for s in sections]}")
    return None


def get_page_url(course: Course, page_name: str):
    pages = course.get_pages()
    for page in pages:
        if page.title == page_name:
            return page.url
    print_red(f"Could not find page {page_name}")
    return None


def get_quiz(course: Course, title: str, delete=False):
    return get_canvas_object(course.get_quizzes, "title", title, delete)


def get_assignment(course: Course, assignment_name, delete=False):
    return get_canvas_object(course.get_assignments, "name", assignment_name, delete)


def get_module(course: Course, module_name: str, delete=False):
    return get_canvas_object(course.get_modules, "name", module_name, delete)


def get_canvas_object(course_getter, attr_name, attr, delete=False):
    objects = course_getter()
    for obj in objects:
        if obj.__getattribute__(attr_name) == attr:
            if not delete:
                return obj
            obj.delete()
    return None


def get_override(assignment: Assignment, override_name, delete=False):
    return get_canvas_object(assignment.get_overrides, "title", override_name, delete)


def get_page(course: Course, title, delete=False):
    return get_canvas_object(course.get_pages, "title", title, delete)


def get_module_item(module: Module, item_name, delete=False):
    return get_canvas_object(module.get_module_items, "title", item_name, delete)


def get_object_id_from_element(course: Course, item):
    if item["type"] == "Quiz":
        if not (quiz := get_quiz(course, item["title"])):
            return None
        return quiz.id
    elif item["type"] == "Assignment":
        if not (assignment := get_assignment(course, item["title"])):
            return None
        return assignment.id
    elif item["type"] == "Page":
        page_url = get_page_url(course, item["title"])
        item["page_url"] = page_url


def fix_dates_attribute(element, attribute, time_zone):
    if attribute in element:
        datetime_version = datetime.fromisoformat(make_iso(element[attribute], time_zone))
        utc_version = datetime_version.astimezone(pytz.utc)
        element[attribute] = utc_version.isoformat()


def fix_dates(element, time_zone):
    print(f"Adding time zone to dates ... ", end="")
    fix_dates_attribute(element, "due_at", time_zone)
    fix_dates_attribute(element, "unlock_at", time_zone)
    fix_dates_attribute(element, "lock_at", time_zone)
    fix_dates_attribute(element, "show_correct_answers_at", time_zone)


def modify_assignment(course, element, delete: bool):
    name = element["name"]
    if canvas_assignment := get_assignment(course, name, delete):
        print(f"Editing canvas assignment {name} ...  ", end="")
        canvas_assignment.edit(assignment=element["settings"])
    else:
        print(f"Creating canvas assignment {name} ...  ", end="")
        course.create_assignment(assignment=element["settings"])
    print("Done")
    return canvas_assignment


def modify_quiz(course: Course, element, delete: bool):
    name = element["name"]
    settings: dict = element["settings"]
    print("Getting quiz from canvas")
    if canvas_quiz := get_quiz(course, name, delete):
        canvas_quiz: Quiz
        print(f"Editing canvas quiz {name} ...  ", end="")
        canvas_quiz.edit(quiz=element["settings"])
    else:
        canvas_quiz = create_quiz(course, element, name, settings)

    replace_questions(canvas_quiz, element["questions"])
    canvas_quiz.edit()
    print("Done")
    return canvas_quiz


def create_quiz(course, element, name, settings):
    print(f"Creating canvas quiz {name} ...  ", end="")
    try:
        canvas_quiz = course.create_quiz(quiz=element["settings"])
    except Exception as ex:
        print(ex)
        print()
        # Perhaps the quiz was partially created, and then the program crashed
        if canvas_quiz := get_quiz(course, name):
            canvas_quiz: Quiz
            print_red("Attempting to edit partially created quiz ...")
            try:
                canvas_quiz.edit(quiz=settings)
            except Exception as ex:
                print_red("Failed to edit quiz")
                raise ex
        else:
            print_red("Quiz was not created")
            print_red("Attempting to debug quiz creation")
            canvas_quiz = debug_quiz_creation(canvas_quiz, course, settings)
    return canvas_quiz


def debug_quiz_creation(canvas_quiz, course, settings):
    for index in range(1, len(settings)):
        keys = list(settings.keys())[:index]
        values = list(settings.values())[:index]
        print_red(f"Attempting with {dict(zip(keys, values))}")
        try:
            canvas_quiz = course.create_quiz(quiz=dict(zip(keys, values)))
        except Exception as ex:
            print(f"Failed on key: {keys[-1]}, value: {values[-1]}")
            raise ex
        canvas_quiz.delete()
    return canvas_quiz


def upload_and_link_files(document_object, course, resources: list[tuple], course_folders):
    """
    Uploads all the files in the resources list, and replaces the fake ids in the document with the real ids.
    """
    print(f"Uploading resources ... ", end="")
    create_resource_folder(course, document_object["name"], course_folders)
    text = json.dumps(document_object, indent=4)
    for fake_id, full_path in resources:
        try:
            resource_id = str(course.upload(full_path)[1]["id"])
            text = text.replace(fake_id, resource_id)
        except IOError:
            print(f"File: {full_path} does not exist, leaving as is")
    return json.loads(text)


def delete_module_item_if_exists(module, name):
    for item in module.get_module_items():
        if item.title == name:
            print(f"Deleting module item {name} ...")
            item.delete()


def create_or_edit_module_item(module: Module, element, object_id, position):
    """
    Creates a module item with an object id, like an assignment or a quiz.
    """
    element["position"] = position
    if not object_id:
        create_or_edit_module_item_without_id(module, element)
    else:
        create_or_edit_module_item_with_id(module, element, object_id)


def create_or_edit_module_item_with_id(module: Module, element, object_id):
    """
    Create module item if it doesn't exist, otherwise edit it.
    """
    element["content_id"] = object_id
    if module_item := get_module_item(module, element["title"]):
        print(f"Editing module item {element['title']} in module {module.name} ...  ", end="")
        module_item.edit(module_item=element)
    else:
        print(f"Creating module item {element['title']} in module {module.name} ...  ", end="")
        module.create_module_item(module_item=element)
    print("Done")


def create_or_edit_module_item_without_id(module: Module, element):
    """
    Creates a module item without an object id, like a page or a header.
    """
    if element["type"] not in ["ExternalUrl", "SubHeader", "Page"]:
        print_red(f"{element['title']} does not exist, no id found when creating module.")
        return

    for item in module.get_module_items():
        if item.title == element["title"]:
            print(f"Editing module item {element['title']} in module {module.name} ...  ", end="")
            item.edit(module_item=element)
            print("Done")
            return

    if element["type"] == "Page" and not element["page_url"]:
        print_red(f"Could not find page url for {element['title']}")
        return

    print(f"Creating module item {element['title']} in module {module.name} ...  ", end="")
    module.create_module_item(module_item=element)
    print("Done")


def delete_module_items_from_element(canvas_module, element):
    names = []
    for item in element["items"]:
        names.append(item["title"])
    for item in canvas_module.get_module_items():
        if item.title not in names:
            print(f"Deleting module item {item.title} ...")
            item.delete()


def create_or_update_module_items(course: Course, element, canvas_module):
    if "items" not in element:
        return
    delete_module_items_from_element(canvas_module, element)
    for index, item in enumerate(element["items"]):
        object_id = get_object_id_from_element(course, item)
        create_or_edit_module_item(canvas_module, item, object_id, index + 1)


def modify_module(course, element):
    name = element["name"]
    if canvas_module := get_module(course, name):
        print(f"Editing canvas module {name} ...  ", end="")
        canvas_module.edit(module=element["settings"])
    else:
        print(f"Creating canvas module {name} ...  ", end="")
        canvas_module = course.create_module(module=element["settings"])
    print()
    create_or_update_module_items(course, element, canvas_module)
    return canvas_module


def get_assignment_override_pairs(course, overrides):
    """
    Searches for canvas assignments with names that match the override names.
    """
    assignments = course.get_assignments()
    pairs = []
    for assignment in assignments:
        for override in overrides:
            if assignment.name == override["title"]:
                pairs.append((assignment, override))
    return pairs


def create_or_update_override_for_assignment(assignment, override, students, sections, section_ids):
    overrides = []
    if students:
        student_override = override.copy()
        student_override["student_ids"] = students
        student_override["title"] = "".join(students)
        overrides.append(student_override)

    for section, s_id in zip(sections, section_ids):
        section_override = override.copy()
        section_override["title"] = section
        section_override["course_section_id"] = s_id
        overrides.append(section_override)

    for override in overrides:
        if canvas_override := get_override(assignment, override["title"]):
            print(f"Editing override {override['title']} ...  ", end="")
            canvas_override.edit(assignment_override=override)
        else:
            print(f"Creating override {override['title']} ...  ", end="")
            assignment.create_override(assignment_override=override)
        print("Done")


def modify_override(course, override, time_zone: str):
    students = override["students"]
    sections = override["sections"]
    section_ids = get_section_ids(course, sections)
    assignment_names = [a['title'] for a in override["assignments"]]

    assignment_override_pairs = get_assignment_override_pairs(course, override["assignments"])
    if not assignment_override_pairs:
        print_red(f"Could not find {assignment_names} in canvas for override {override['sections']}")
        return
    if not students and not sections:
        raise ValueError("Must provide either students or sections")

    for assignment, override in assignment_override_pairs:
        fix_dates(override, time_zone)
        create_or_update_override_for_assignment(assignment, override, students, sections, section_ids)


def get_section_ids(course, names):
    sections = course.get_sections()
    sections = [s.id for s in sections if s.name in names]
    if not sections:
        raise ValueError(f"Could not find sections {sections}")
    return sections


def modify_page(course: Course, element, delete: bool):
    name = element["name"]
    if canvas_page := get_page(course, name, delete):
        print(f"Editing canvas page {name} ...  ", end="")
        canvas_page.edit(wiki_page=element["settings"])
    else:
        print(f"Creating canvas page {name} ...  ", end="")
        canvas_page = course.create_page(wiki_page=element["settings"])
    print("Done")
    return canvas_page


def post_document(course: Course, time_zone, file_path: Path, delete: bool = False):
    """
    Parses a markdown file, and posts the elements to Canvas.
    :param course: The canvas course to post to, obtained from the canvas api
    :param time_zone: The time zone of the course (e.g. "America/Denver")
    :param file_path: The path to the markdown file
    :param delete: If true, deletes all elements in the Canvas course with the same name as the elements in the file
    """

    print(f"Parsing file ({file_path.name}) ...  ", end="")
    if "mdx" not in file_path.name:
        print_red("Error: File must be a mdx file")
        return

    assignment_groups = list(course.get_assignment_groups())
    names_to_ids = {g.name: g.id for g in assignment_groups}

    # Provide processing functions, so that the parser needs no access to a canvas course
    parser = DocumentParser(
        path_to_resources=file_path.parent,
        path_to_canvas_files=file_path.parent,
        markdown_processor=lambda text: process_markdown(text, course, file_path.parent),
        time_zone=time_zone,
        group_identifier=lambda group_name: get_group_id(course, group_name, names_to_ids),
    )
    document_object = parser.parse(file_path.read_text())

    course_folders = list(course.get_folders())

    # Create multiple quizzes or assignments from the document object
    for element in document_object:
        if "resources" in element:
            element = upload_and_link_files(element, course, element["resources"], course_folders)
        if "settings" in element:
            fix_dates(element["settings"], time_zone)
        if element["type"] == "quiz":
            modify_quiz(course, element, delete)
        elif element["type"] == "assignment":
            modify_assignment(course, element, delete)
        elif element["type"] == "page":
            modify_page(course, element, delete)
        elif element["type"] == "module":
            modify_module(course, element)
        elif element["type"] == "override":
            modify_override(course, element, time_zone)
        else:
            raise ValueError(f"Unknown type {element['type']}")


def get_course(api_url: str, api_token: str, canvas_course_id: int) -> Course:
    """
    Returns a Canvas Course object for the given API URL, API token, and course ID.

    :param api_url: str: The URL for the Canvas API.
    :param api_token: str: The authentication token for the Canvas API.
    :param canvas_course_id: int: The ID of the Canvas course.
    :return: Course: A Canvas Course object.
    """
    canvas = Canvas(api_url, api_token)
    course: Course = canvas.get_course(canvas_course_id)
    return course
